import os
from termcolor import cprint
from .langchain_tools import build_langchain_tools, is_langchain_installed
from bondai.tools import DalleTool, PythonREPLTool, ShellTool
from bondai.tools.alpaca_markets import CreateOrderTool, GetAccountTool, ListPositionsTool
from bondai.tools.file import FileQueryTool, FileWriteTool
from bondai.tools.gmail import ListEmailsTool, QueryEmailsTool
from bondai.tools.search import GoogleSearchTool, DuckDuckGoSearchTool
from bondai.tools.database import DatabaseQueryTool
from bondai.tools.bland_ai import BlandAITool
from bondai.tools.website import (
    DownloadFileTool,
    WebsiteQueryTool,
)

def get_tools():
    tool_options = [
        DownloadFileTool(),
        FileQueryTool(),
        FileWriteTool(), 
        WebsiteQueryTool(),
        DalleTool(),
        PythonREPLTool(),
        ShellTool(),
    ]

    if os.environ.get('ALPACA_MARKETS_API_KEY') and os.environ.get('ALPACA_MARKETS_SECRET_KEY'):
        tool_options.append(CreateOrderTool())
        tool_options.append(GetAccountTool())
        tool_options.append(ListPositionsTool())
    else:
        cprint("Skipping Alpaca Markets tools because ALPACA_MARKETS_API_KEY and ALPACA_MARKETS_SECRET_KEY environment variables are not set.", "yellow")

    if os.environ.get('GOOGLE_API_KEY') and os.environ.get('GOOGLE_CSE_ID'):
        tool_options.append(GoogleSearchTool())
    else:
        tool_options.append(DuckDuckGoSearchTool())
        cprint("Skipping Google Search tool because GOOGLE_API_KEY and GOOGLE_CSE_ID environment variables are not set.", "yellow")

    if os.environ.get('BLAND_AI_API_KEY'):
        tool_options.append(BlandAITool())
    else:
        cprint("Skipping Bland AI tool because BLAND_AI_API_KEY environment variable is not set.", "yellow")

    if os.environ.get('PG_URI') or os.environ.get('PG_HOST'):
        tool_options.append(DatabaseQueryTool())
    else:
        cprint("Skipping Database tools because PG_URI and PG_HOST environment variables are not set. One of these must be set to enable Database connectivity.", "yellow")

    if 'gmail-token.pickle' in os.listdir():
        tool_options.append(ListEmailsTool())
        tool_options.append(QueryEmailsTool())
    else:
        cprint("Skipping Gmail tools because gmail-token.pickle file is not present.", "yellow")

    return tool_options




import threading
import subprocess
import shlex
from queue import Queue
from pydantic import BaseModel
from bondai.tools import Tool

TOOL_NAME = 'shell_tool'
TOOL_DESCRIPTION = (
    "This tool allows you to execute shell commands. "
    "Specify your command in the 'command' parameter and it will return the result. "
    "Note that this tool only accepts a single string argument ('command') at a time and does not accept a list of commands."
)

class Parameters(BaseModel):
    command: str
    thought: str

class ShellTool(Tool):
    def __init__(self):
        super(ShellTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, parameters=Parameters, dangerous=True)

    def run(self, arguments):
        cmd = arguments.get('command')
        if cmd is None:
            raise Exception("'command' parameter is required")

        stdout, stderr = self.execute_command(cmd)
        
        response = ""
        
        # Include stdout if present
        if stdout:
            response += f"Output:\n{stdout}\n"
        
        # Include stderr if present
        if stderr:
            response += f"Errors:\n{stderr}\n"
        
        if not response:
            response = "Command executed successfully. No output."

        return response

    def execute_command(self, cmd):
        # Use threading to enforce timeout
        def target(queue):
            process = subprocess.Popen(
                shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            stdout, stderr = process.communicate()
            queue.put((stdout, stderr))

        q = Queue()
        thread = threading.Thread(target=target, args=(q,))
        thread.start()
        thread.join(timeout=10)  # 10 seconds timeout
            
        if thread.is_alive():
            thread.join()  # Ensure it's stopped
            raise Exception("Command execution timed out")

        stdout, stderr = q.get()  # Get the result from the queue

        return stdout.decode('utf-8'), stderr.decode('utf-8')



from .client import BondAIAPIClient
from .conversation_tool import ConversationTool
from .server import BondAIAPIServer, AgentWrapper, BondAIAPIError

__all__ = [
    'BondAIAPIClient',
    'BondAIAPIServer',
    'AgentWrapper',
    'ConversationTool',
    'BondAIAPIError',
]


import json
import threading
from bondai.tools import Tool, InputParameters

TOOL_NAME = 'conversation_tool'
TOOL_DESCRIPTION = (
    "This tool allows you to communicate with the user and ask them questions."
    "To use this tool just put your message in the 'input' parameter."
    "Remember to always be friendly and polite!"
)

class ConversationTool(Tool):
    def __init__(self, socketio):
        super(ConversationTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, InputParameters)
        self.socketio = socketio
        self.message_arrived_event = threading.Event()
        self.user_message = None
        self._setup_socket_listener()

    def _setup_socket_listener(self):
        # Set up the event listener once during initialization
        @self.socketio.on('message')
        def handle_message(message):
            message = json.loads(message)
            if message.get('event') == 'user_message':
                self.user_message = message['data']['message']
                self.message_arrived_event.set()

    def run(self, arguments):
        # Check for required arguments
        question = arguments.get('input')
        if not question:
            raise ValueError("'input' argument is required")

        # Reset for each run
        self.user_message = None
        self.message_arrived_event.clear()

        # Emit message, now that our listener is guaranteed to be active
        message = {
            'event': 'agent_message',
            'data': {
                'message': question
            }
        }
        payload = json.dumps(message)
        self.socketio.send(payload)

        # Wait for user message
        self.message_arrived_event.wait()
        result = self.user_message
        self.user_message = None
        print(result)

        return result



import json
import threading
from pydantic import BaseModel
import traceback
from termcolor import colored, cprint
from bondai.tools import Tool, ResponseQueryTool, AgentTool
from bondai.prompt import DefaultPromptBuilder
from bondai.models.openai import (
    OpenAILLM, 
    MODEL_GPT4_0613, 
    MODEL_GPT35_TURBO_0613, 
    get_total_cost
)

class FinalAnswerParameters(BaseModel):
    input: str

TOOL_MAX_TOKEN_RESPONSE = 2000
DEFAULT_FINAL_ANSWER_TOOL = Tool('final_answer', "Use the final_answer tool when you have all the information you need to provide the final answer.", FinalAnswerParameters)

AGENT_STATE_RUNNING = 'AGENT_STATE_RUNNING'
AGENT_STATE_STOPPED = 'AGENT_STATE_STOPPED'

def format_print_string(s, length=100):
    # Remove newlines
    s = s.replace('\n', ' ').replace('\r', '')
    
    if len(s) <= length:
        return s
    return s[:length - 3] + "..."


class AgentStep:

    def __init__(self, prompt, message, function=None):
        self.prompt = prompt
        self.message = message
        self.function = function
        self.output = None
        self.error = False
        self.exit = False

class BudgetExceededException(Exception):
    pass

class MaxStepsExceededException(Exception):
    pass

class Agent:

    def __init__(self, prompt_builder=None, tools=[], llm=OpenAILLM(MODEL_GPT4_0613), fallback_llm=OpenAILLM(MODEL_GPT35_TURBO_0613), final_answer_tool=DEFAULT_FINAL_ANSWER_TOOL, quiet=False, enable_sub_agent=False):
        if not prompt_builder:
            self.prompt_builder = DefaultPromptBuilder(llm)
        else:
            self.prompt_builder = prompt_builder
        
        if enable_sub_agent:
            tools.append(AgentTool(tools=tools, llm=llm))
        
        self.state = AGENT_STATE_STOPPED
        self.previous_steps = []
        self.previous_messages = []
        self.response_query_tool = ResponseQueryTool()
        self.tools = tools
        self.llm = llm
        self.fallback_llm = fallback_llm
        self.final_answer_tool = final_answer_tool
        self.quiet = quiet
        self._thread = None
        self._stop_thread = False
        self._events = {
            'started': [],
            'step_completed': [],
            'completed': [],
        }
    
    def on(self, event_name):
        """Register a callback to an event."""
        if event_name not in self._events:
            raise ValueError(f"Unsupported event '{event_name}'")

        def decorator(callback):
            self._events[event_name].append(callback)
            return callback
        
        return decorator
    
    def _trigger_event(self, event_name, *args, **kwargs):
        """Trigger the specified event."""
        for callback in self._events.get(event_name, []):
            callback(*args, **kwargs)


    def add_tool(self, tool):
        if self.state == AGENT_STATE_RUNNING:
            raise Exception('Cannot modify agent while it is in a running state.')
        if not any([t.name == tool.name for t in self.tools]):
            self.tools.append(tool)
    
    def remove_tool(self, tool_name):
        if self.state == AGENT_STATE_RUNNING:
            raise Exception('Cannot modify agent while it is in a running state.')
        self.tools = [t for t in self.tools if t.name != tool_name]
    
    def run_once(self, task=''):
        tools = self.tools
        if len(self.response_query_tool.responses) > 0:
            tools.append(self.response_query_tool)

        prompt = self.prompt_builder.build_prompt(task, tools, self.previous_steps)
        functions = list(map(lambda t: t.get_tool_function(), tools))

        use_streaming = self.llm.supports_streaming() and any([t.supports_streaming for t in self.tools])
        if use_streaming:
            def function_stream_callback(function_name, arguments_buffer):
                streaming_tools = [t for t in self.tools if t.name == function_name and t.supports_streaming]
                if len(streaming_tools) > 0:
                    tool = streaming_tools[0]
                    tool.handle_stream_update(arguments_buffer)
            
            message, function = self.llm.get_streaming_completion(
                prompt, 
                previous_messages=self.previous_messages, 
                functions=functions, 
                function_stream_callback=function_stream_callback
            )
        else:
            message, function = self.llm.get_completion(
                prompt, 
                previous_messages=self.previous_messages, 
                functions=functions
            )

        if not function:
            message, function = self.fallback_llm.get_completion(
                message, 
                system_prompt='Select the correct function and parameters to use based on the prompt', 
                previous_messages=[], 
                functions=functions
            )

        if not function and not self.quiet:
            cprint(f"No function was selected.", 'red')
            cprint(message, 'red')
            
        if function:
            step = AgentStep(prompt, message, function)
            tools = [t for t in self.tools if t.name == function['name']]
            if len(tools) > 0:
                tool = tools[0]
                
                try:
                    args = function['arguments'] if 'arguments' in function else {}
                    
                    if not self.quiet:
                        cprint(f"\n\nUsing the {tool.name} tool", 'yellow', attrs=["bold"])
                        if len(args) > 0:
                            if 'thought' in args:
                                print(colored("Thought:", 'white', attrs=["bold"]), colored(args['thought'], 'white'))

                            cprint("Arguments", 'white', attrs=["bold"])
                            for key, value in args.items():
                                if key != 'thought':
                                    cprint(f"{key}: {format_print_string(str(value))}", 'white')


                    try:
                        step.output = tool.run(args)
                        step.exit = tool.exit_agent or (self.final_answer_tool and tool.name == self.final_answer_tool.name)
                        
                        if step.output:
                            if self.llm.count_tokens(step.output) > TOOL_MAX_TOKEN_RESPONSE:
                                response_id = self.response_query_tool.add_response(step.output)
                                step.output = f"The result from this tool was greater than {TOOL_MAX_TOKEN_RESPONSE} tokens and could not be displayed. However, you can use the response_query tool to ask questions about the content of this response. Just use response_id = {response_id}."
                        else:
                            step.output = "This command ran successfully with no output."
                        
                        if not self.quiet:
                            print(colored("Output:", 'white', attrs=["bold"]), colored(format_print_string(step.output), 'white'))
                    except Exception as e:
                        # traceback.print_exc()
                        if not self.quiet:
                            cprint(f"An Error occured: {e}", 'red', attrs=["bold"])
                        step.error = True
                        step.output = "An Error occured: " + str(e)
                except Exception as e:
                    if not self.quiet:
                        cprint(f"An Error occured: {e}", 'red', attrs=["bold"])
                    step.error = True
                    step.output = 'An Error occured: The provided arguments were not valid JSON. Valid JSON must ALWAYS BE PROVIDED in the response.'
            else:
                step.error = True
                step.output = f"An Error occured: The function '{function['name']}' does not exist."

                if not self.quiet:
                    cprint(f"No tools found for: {function['name']}", 'red', attrs=["bold"])
                
        else:
            step = AgentStep(prompt, message)
            step.output = "No function was provided."
        
        return step
    
    def reset_memory(self):
        if self.state == AGENT_STATE_RUNNING:
            raise Exception('Cannot modify agent while it is in a running state.')
        self.previous_steps = []

    def run(self, task='', task_budget=None, max_steps=None):
        task = '' if not task else task
        if self.state == AGENT_STATE_RUNNING:
            raise Exception('Cannot modify agent while it is in a running state.')
        
        self.state = AGENT_STATE_RUNNING
        try:
            self._trigger_event('started')
            if self.final_answer_tool:
                self.tools = self.tools + [self.final_answer_tool]

            step_counter = 0
            while True:
                if self._stop_thread:
                    break
                step_counter += 1
                step = self.run_once(task)

                if step.function or len(self.previous_steps) == 0 or self.previous_steps[-1].function:
                    self.previous_steps.append(step)
                    self._trigger_event('step_completed', step)
                
                total_cost = get_total_cost()
                if not self.quiet:
                    print(colored("Total Cost:", 'green', attrs=["bold"]), colored('$' + str(round(total_cost, 2)), 'white'))

                if step.exit:
                    self.previous_messages.append({
                        'prompt': task,
                        'response': step.output
                    })
                    return step

                if task_budget and total_cost >= task_budget:
                    raise BudgetExceededException("The budget for this task has been reached without successfully finishing.")
                if max_steps and step_counter >= max_steps:
                    raise MaxStepsExceededException("The maximum number of steps has been reached without successfully finishing.")
        finally:
            self._trigger_event('completed')
            self.state = AGENT_STATE_STOPPED

    def run_async(self, task='', task_budget=None, max_steps=None):
        """Runs the agent's task in a separate thread."""
        if self.state == AGENT_STATE_RUNNING or (self._thread and self._thread.is_alive()):
            raise Exception('Cannot start agent while it is in a running state.')
        self._thread = threading.Thread(target=self.run, args=(task, task_budget, max_steps))
        self._thread.start()

    def join(self):
        """Blocks until the thread completes."""
        if self._thread:
            self._thread.join()

    def stop(self):
        """Gracefully stops the thread."""
        self._stop_thread = True
        if self._thread:
            self._thread.join()
        self._stop_thread = False




from .agent import Agent, AGENT_STATE_RUNNING, AGENT_STATE_STOPPED, BudgetExceededException, MaxStepsExceededException

__all__ = [
    "Agent",
    "BudgetExceededException",
    "MaxStepsExceededException",
    "AGENT_STATE_RUNNING",
    "AGENT_STATE_STOPPED",
]


import threading
import io
import sys
from contextlib import redirect_stdout, redirect_stderr
from pydantic import BaseModel
from bondai.tools import Tool

TOOL_NAME = 'python_repl'
TOOL_DESCRIPTION = (
    "This tool allows you to execute Python code. "
    "Specify your Python code in the 'code' parameter and it will return the result."
)

class Parameters(BaseModel):
    code: str
    thought: str

class PythonREPLTool(Tool):
    def __init__(self, execution_timeout=30):
        super(PythonREPLTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, parameters=Parameters, dangerous=True)
        self.execution_timeout = execution_timeout
    
    def run(self, arguments):
        code = arguments.get('code')

        if code is None:
            raise Exception("'code' parameter is required")

        result, stdout, stderr = self.execute_code(code)

        print("result:", result)
        print("stdout:", stdout)
        print("stderr:", stderr)
        
        response = ""
        
        # Include stdout if present
        if stdout:
            response += f"Output:\n{stdout}\n"
            
        # Include stderr if present
        if stderr:
            response += f"Errors:\n{stderr}\n"
        
        # Include result if present
        if result:
            formatted_result = "\n".join([f"{key}: {value}" for key, value in result.items()])
            response += f"Result Variables:\n{formatted_result}\n"

        if not response:
            response = "Code executed successfully. No output or result variables."
        
        return response
    
    def execute_code(self, code):
        # Capture stdout and stderr
        thread_exception = None
        stdout_io = io.StringIO()
        stderr_io = io.StringIO()

        # Use threading to enforce timeout
        def target(local_vars, code):
            nonlocal thread_exception
            try:
                with redirect_stdout(stdout_io), redirect_stderr(stderr_io):
                    exec(code, {}, local_vars)
            except Exception as e:
                thread_exception = e
        
        local_vars = {}
        thread = threading.Thread(target=target, args=(local_vars, code))
        thread.start()
        thread.join(timeout=self.execution_timeout)

        if thread_exception:
            raise thread_exception
        
        if thread.is_alive():
            thread.join()  # Ensure it's stopped
            raise Exception("Code execution timed out")

        stdout = stdout_io.getvalue()
        stderr = stderr_io.getvalue()

        return local_vars, stdout, stderr




import json
from flask import Flask, request, jsonify
from flask_cors import CORS
from flask_restful import Api, Resource
from flask_socketio import SocketIO

from bondai import AGENT_STATE_RUNNING

class BondAIAPIError(Exception):
    pass

class BondAIAPIServer:
    def __init__(self, agent=None, tools=None, agent_wrapper=None, port=2663):
        if agent_wrapper:
            self.agent_wrapper = agent_wrapper
        elif agent and tools:
            self.agent_wrapper = AgentWrapper(agent, tools)
        else:
            raise Exception('Either agent_wrapper or agent and tools must be provided.')
        
        self.port = port
        self.app = Flask(__name__)
        self.api = Api(self.app)
        self.socketio = SocketIO(self.app)
        CORS(self.app)  # Enable CORS for the Flask app

        self._setup_resources()
        self._setup_agent_events()

    def _setup_resources(self):
        self.api.add_resource(AgentResource, '/agent', resource_class_args=(self.agent_wrapper,))
        self.api.add_resource(StartAgentResource, '/agent/start', resource_class_args=(self.agent_wrapper,))
        self.api.add_resource(AgentToolsResource, '/agent/tools', '/agent/tools/<string:tool_name>', resource_class_args=(self.agent_wrapper,))
        self.api.add_resource(ToolsListResource, '/tools', resource_class_args=(self.agent_wrapper,))

    def _setup_agent_events(self):
        @self.agent_wrapper.agent.on('started')
        def handle_agent_started():
            data = { 'event': 'agent_started' }
            payload = json.dumps(data)
            self.socketio.send(payload)
        
        @self.agent_wrapper.agent.on('completed')
        def handle_agent_completed():
            data = { 'event': 'agent_completed' }
            payload = json.dumps(data)
            self.socketio.send(payload)
        
        @self.agent_wrapper.agent.on('step_completed')
        def handle_agent_step_completed(step):
            step_data = step.__dict__
            del step_data['prompt']
            data = {
                'event': 'agent_step_completed',
                'data': {
                    'step': step_data
                }
            }
            payload = json.dumps(data)
            self.socketio.send(payload)

    def run(self):
        self.socketio.run(self.app, port=self.port)

    def shutdown(self):
        # Use this function to gracefully shutdown any resources if needed
        print("Shutting down BondAIAPI...")
        self.socketio.stop()

class AgentWrapper:
    def __init__(self, agent, tools):
        self.agent = agent
        self.tools = tools
    
    def find_tool(self, tool_name):
        for tool in self.tools:
            if tool.name == tool_name:
                return tool
        return None
    
    def get_previous_steps(self):
        return [s.__dict__ for s in self.agent.previous_steps]

    def get_tools(self):
        return [t.get_tool_function() for t in self.tools]

    def get_agent(self):
        agent_tools = [t.get_tool_function() for t in self.agent.tools]
        return {
            'state': self.agent.state,
            'previous_steps': self.get_previous_steps(),
            'previous_messages': self.agent.previous_messages,
            'tools': agent_tools,
        }

    def start_agent(self, task=None, task_budget=None, max_steps=None):
        if self.agent.state == AGENT_STATE_RUNNING:
            raise BondAIAPIError('Agent cannot be modified when it is already running.')
        self.agent.run_async(task, task_budget=task_budget, max_steps=max_steps)

    def add_tool(self, tool_name):
        if self.agent.state == AGENT_STATE_RUNNING:
            raise BondAIAPIError('Agent cannot be modified when it is already running.')
        
        selected_tool = self.find_tool(tool_name)
        if not selected_tool:
            raise BondAIAPIError(f"Tool '{tool_name}' does not exist.")
        
        if not any([t.name == tool_name for t in self.agent.tools]):
            self.agent.add_tool(selected_tool)

    def remove_tool(self, tool_name):
        if self.agent.state == AGENT_STATE_RUNNING:
            raise BondAIAPIError('Agent cannot be modified when it is already running.')
        self.agent.remove_tool(tool_name)


class StartAgentResource(Resource):
    def __init__(self, agent_wrapper):
        self.agent_wrapper = agent_wrapper

    def post(self):
        data = request.get_json()
        task = data.get('task')
        task_budget = data.get('task_budget')
        max_steps = data.get('max_steps')

        try:
            self.agent_wrapper.start_agent(task=task, task_budget=task_budget, max_steps=max_steps)
        except BondAIAPIError as e:
            return str(e), 400
        
        return jsonify({'status': 'success'})

class AgentResource(Resource):
    def __init__(self, agent_wrapper):
        self.agent_wrapper = agent_wrapper

    def get(self):
        try:
            return jsonify(self.agent_wrapper.get_agent())
        except BondAIAPIError as e:
            return str(e), 400

class AgentToolsResource(Resource):
    def __init__(self, agent_wrapper):
        self.agent_wrapper = agent_wrapper

    def post(self):
        data = request.get_json()
        tool_name = data['tool_name']
        if not tool_name:
            return 'tool_name is required.', 400

        try:
            self.agent_wrapper.add_tool(tool_name)
        except BondAIAPIError as e:
            return str(e), 400
        
        return jsonify({'status': 'success'})

    def delete(self, tool_name):
        if not tool_name:
            return jsonify({'error': 'tool_name is required.'}), 400

        try:
            self.agent_wrapper.remove_tool(tool_name)
        except BondAIAPIError as e:
            return str(e), 400

        return jsonify({'status': 'success'})

class ToolsListResource(Resource):
    def __init__(self, agent_wrapper):
        self.agent_wrapper = agent_wrapper

    def get(self):
        return jsonify(self.agent_wrapper.get_tools())



from bondai import AGENT_STATE_RUNNING
from bondai.api import AgentWrapper, BondAIAPIError

class CLIAgentWrapper(AgentWrapper):
    def __init__(self, conversational_agent, task_agent, tools):
        super(CLIAgentWrapper, self).__init__(task_agent, tools)
        self.conversational_agent = conversational_agent

    def get_agent(self):
        agent_tools = [t.get_tool_function() for t in self.agent.tools]
        return {
            'state': self.agent.state,
            'previous_steps': self.get_previous_steps(),
            'previous_messages': self.conversational_agent.previous_messages,
            'tools': agent_tools,
        }
    
    def start_agent(self, task=None, task_budget=None, max_steps=None):
        if self.conversational_agent.state == AGENT_STATE_RUNNING:
            raise BondAIAPIError('Agent cannot be modified when it is already running.')
        self.conversational_agent.run_async(task, task_budget=task_budget, max_steps=max_steps)



from pydantic import BaseModel
from langchain.tools import ShellTool
from langchain.tools.python.tool import PythonREPLTool
from bondai.models.openai import OpenAILLM, MODEL_GPT4_0613
from bondai.tools import Tool, LangChainTool
from bondai.util import load_local_resource
from bondai.prompt import DefaultPromptBuilder

TOOL_NAME = 'tool_maker_tool'
TOOL_DESCRIPTION = (
    "The ToolMaker is a state-of-the-art code generation tool designed to transform natural language tool descriptions into tools that you can use for your work. "
    "The ToolMaker is a great way to store solutions to problems so that you can use them later. "
    "To use the ToolMaker, simply include a highly detailed description of the tool you would like to create in the 'tool_description' parameter."
    "Your 'tool_description' (required) should include a highly detailed description of the desired tool, its name, parameters it should accept, what response it should return, and it's expected behavior. "
    "You should also include a 'code_snippet' (required) parameter that contains examples of code that can be used to create this tool. "
    "Including code snippets will help the ToolMaker be better at creating the tool you want. "
)
PROMPT_TEMPLATE = load_local_resource(__file__, 'prompt_template.md')

class ShellToolParameters(BaseModel):
    commands: str
    thought: str

class PythonREPLParameters(BaseModel):
    query: str

class Parameters(BaseModel):
    tool_description: str
    code_snippet: str
    thought: str

class ToolMakerTool(Tool):
    def __init__(self, llm=OpenAILLM(MODEL_GPT4_0613)):
        super(ToolMakerTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.llm = llm
    
    def run(self, arguments):
        from bondai import Agent
        tool_description = arguments.get('tool_description')
        code_snippet = arguments.get('code_snippet')

        if not tool_description:
            raise ValueError('tool_description is required')
        if not code_snippet:
            raise ValueError('code_snippet is required')

        shell_tool = ShellTool()
        shell_tool.description = "This is a very powerful tool that allows you to run any shell command on this Ubuntu machine. Note that this tool only accepts a single string argument at a time and does not accept a list of commands." + f" args {shell_tool.args}".replace(

            "{", "{{"
        ).replace("}", "}}")

        result = Agent(
            llm=self.llm,
            prompt_builder=DefaultPromptBuilder(llm=self.llm, prompt_template=PROMPT_TEMPLATE), 
            tools=[
                LangChainTool(shell_tool, ShellToolParameters),
                LangChainTool(PythonREPLTool(), PythonREPLParameters)
            ], 
        ).run()
        
        print(result)
        return result.output



from pydantic import BaseModel
from bondai.tools import Tool

TOOL_NAME = 'file_write'
TOOL_DESCRIPTION = (
    "This tool will save the data you provide in the 'text' parameter of this tool to a file."
    "You MUST specify the filename of the file you want to save using the 'filename' parameter."
    "You can optionally specify the 'append' parameter to append the 'text' to the file instead of overwriting it."
)

class Parameters(BaseModel):
    filename: str
    text: str
    append: bool = False
    thought: str

class FileWriteTool(Tool):
    def __init__(self):
        super(FileWriteTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
    
    def run(self, arguments):
        filename = arguments.get('filename')
        text = arguments.get('text')

        if filename is None:
            raise Exception('filename is required')
        if text is None:
            raise Exception('text is required')

        mode = 'a' if arguments.get('append') else 'w'
        with open(filename, mode) as f:
            f.write(text)
            return f"File {filename} written successfully"




from bondai.tools import Tool, InputParameters
from bondai.models.openai import OpenAILLM, MODEL_GPT4_0613

TOOL_NAME = 'agent_tool'
TOOL_DESCRIPTION = (
    "This tool allows you to delegate tasks to other agents. "
    "This can be really helpful for taking a complex task and breaking it down into smaller, more manageable pieces. "
    "Just include a highly descriptive prompt in the 'input' parameter for this task. "
    "The more detailed your description the better the agent will be at the task. "
)

class AgentTool(Tool):
    def __init__(self, agent=None, prompt_builder=None, tools=[], llm=OpenAILLM(MODEL_GPT4_0613)):
        super(AgentTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, InputParameters)
        
        from bondai import Agent
        if agent:
            self.agent = agent
        else:
            self.agent = Agent(prompt_builder=prompt_builder, tools=tools, llm=llm)
    
    def run(self, arguments):
        agent_prompt = arguments['input']
        
        if isinstance(agent_prompt, list):
            output = ''
            for i,p in enumerate(agent_prompt):
                result = self.agent.run(p)
                output += f"Agent #{i+1} Response: {result.output}\n"
            
            return output
        else:
            result = self.agent.run(agent_prompt).output
            self.agent.reset_memory()
            return result



import requests
import openai
from pydantic import BaseModel
from bondai.tools import Tool
from bondai.models.openai.openai_connection_params import DALLE_CONNECTION_PARAMS

IMAGE_SIZE = "1024x1024"
TOOL_NAME = "text_to_image_tool"
TOOL_DESCRIPTION = (
    "This tool takes in a description and generates an image based on that description. "
    "It then saves the image to the specified filename.\n"
    "\nParameters:\n"
    "- description (required): The text description based on which the image will be generated.\n"
    "- filename (required): The name of the file where the generated image will be saved."
)

class Parameters(BaseModel):
    description: str
    filename: str
    thought: str

class DalleTool(Tool):

    def __init__(self):
        super().__init__(TOOL_NAME, TOOL_DESCRIPTION, parameters=Parameters)

    def run(self, arguments):
        description = arguments.get('description')
        filename = arguments.get('filename')

        if description is None:
            raise Exception("description is required.")
        if filename is None:
            raise Exception("filename is required.")
        
        params = {
            'prompt': description,
            'n': 1,
            'size': IMAGE_SIZE
        }

        # Use the OpenAI API to generate an image based on the description
        response = openai.Image.create(
            **params,
            **DALLE_CONNECTION_PARAMS
        )

        # Get the image URL from the response
        image_url = response['data'][0]['url']

        # Download the image from the URL and save it to the specified filename
        image_content = requests.get(image_url).content
        with open(filename, 'wb') as file:
            file.write(image_content)

        return f"Image generated and saved to {filename} successfully."


#!/usr/bin/env python3

import argparse
import os
import json
from termcolor import cprint, colored
from bondai import Agent, BudgetExceededException
from bondai.tools import Tool, HumanTool, AgentTool
from bondai.prompt import DefaultPromptBuilder
from bondai.util import ModelLogger, load_local_resource
from bondai.api import BondAIAPIServer, ConversationTool
from bondai.models.openai import (
    OpenAILLM, 
    MODEL_GPT4_0613, 
    enable_logging,
    OPENAI_CONNECTION_TYPE,
    OPENAI_CONNECTION_TYPE_OPENAI,
)

from .default_tools import get_tools
from .cli_agent_wrapper import CLIAgentWrapper

if OPENAI_CONNECTION_TYPE == OPENAI_CONNECTION_TYPE_OPENAI and not os.environ.get('OPENAI_API_KEY'):
    cprint(f"The OPENAI_API_KEY environment variable has not been set. Please input your OpenAI API Key now or type 'exit'.", 'yellow')
    user_input = input()
    if user_input == 'exit':
        exit(1)
    else:
        import openai
        openai.api_key = user_input




cprint(f"Loading BondAI...", 'white')

parser = argparse.ArgumentParser(description="BondAI CLI tool options")

parser.add_argument('--server', 
                    nargs='?', 
                    const='2663', 
                    metavar='server_port',
                    help='Starts the BondAI web server. If no port is specified, defaults to 5000.')

parser.add_argument('--enable-dangerous', 
                    action='store_true', 
                    help='Allows potentially dangerous Tools to be loaded (i.e. ShellTool and PythonREPLTool)')

# --enable-prompt-logging with optional log_dir
parser.add_argument('--enable-prompt-logging', 
                    nargs='?', 
                    const='logs', 
                    metavar='log_dir',
                    help='Turns on prompt logging which will write all prompt inputs into the specified directory. Defaults to "logs" if no directory provided.')

# --load-tools with specified Python file
parser.add_argument('--load-tools', 
                    metavar='my_tools.py', 
                    help='Specify a Python file to load tools from. The file should have a function named get_tools() that returns a list of Tools.')

 # --quiet
parser.add_argument('--quiet', 
                    action='store_true', 
                    default=False,
                    help='If set, minimizes the output to the console.')

args = parser.parse_args()


if args.enable_prompt_logging:
    log_dir = args.enable_prompt_logging
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    
    enable_logging(ModelLogger(log_dir))
    cprint(f"Prompt logging is enabled. Logs will be written to: {log_dir}", "yellow")

tools = []
if args.load_tools:
    tools_file = args.load_tools
    if os.path.exists(tools_file):
        import importlib.util
        spec = importlib.util.spec_from_file_location("module.name", tools_file)
        tools_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(tools_module)
        tools = tools_module.get_tools()  # Assuming get_tools() returns a list of tools
        cprint(f"Loaded tools from {tools_file}", "yellow")
    else:
        cprint(f"Error: {tools_file} does not exist.", "red")
        exit(1)
else:
    tools = get_tools()

if args.enable_dangerous:
    cprint("Dangerous Tools are enabled.", "red")
else:
    tools = [tool for tool in tools if not tool.dangerous]

tool_descriptions = ''.join([f"{tool.name}: {tool.description}\n" for tool in tools])

onboarding_prompt_template = load_local_resource(__file__, 'onboarding_prompt_template.md')
onboarding_prompt_template = onboarding_prompt_template.replace('{TOOLS}', tool_descriptions)

llm=OpenAILLM(MODEL_GPT4_0613)

task_agent = Agent(llm=llm, tools=tools, quiet=args.quiet, enable_sub_agent=True)
task_agent_tool = AgentTool(task_agent)
task_agent_tool.description = (
    "This tool allows you to use the BondAI Agent to solve the user's task."
    "To use this tool you must include your thoughtful, highly detailed task description in the 'input' parameter."
    "Your task description MUST be highly detailed and include ALL information useful for solving the user's task."
    "The Agent will then attempt to complete the task."
    "You MUST share the results of the Agent's work with the user using the human_tool."
    "The user will not be able to see the Agent's work unless you show them the results using the human_tool."
)

human_tool = HumanTool()
human_tool.description = (
    "This tool allows you to communicate with the user and ask them questions."
    "To use this tool just put your question in the 'input' parameter."
    "Remember to always be friendly and polite!"
)

exit_tool = Tool('exit_tool', (
    "This tool allows you to exit the BondAI CLI."
    'You MUST call this tool if the user wants to exit the application.'
    'Do not use this tool unless the user says they want to exit the application.'
))


def run_cli():
    try:
        if args.server:
            port = int(args.server)
            api_agent = Agent(
                llm = llm,
                prompt_builder=DefaultPromptBuilder(llm, onboarding_prompt_template), 
                tools=[task_agent_tool],
                final_answer_tool=None,
            )
            agent_wrapper = CLIAgentWrapper(api_agent, task_agent, tools)
            server = BondAIAPIServer(agent_wrapper=agent_wrapper, port=port)

            conversational_tool = ConversationTool(server.socketio)
            api_agent.add_tool(conversational_tool)

            server.run()
        else:
            cli_agent = Agent(
                llm = llm,
                prompt_builder=DefaultPromptBuilder(llm, onboarding_prompt_template), 
                final_answer_tool=exit_tool,
                tools=[
                    human_tool,
                    task_agent_tool
                ], 
                quiet=True
            )
            cli_agent.run()
    except BudgetExceededException as e:
        cprint(f"\n\nThe budget for this task has been exceeded and will stop.\n", 'red')



import os
import json
import time
import openai
import tiktoken
from .openai_models import MODELS, MODEL_TYPE_LLM

TEMPERATURE = 0.1

embedding_tokens = 0
embedding_costs = 0.0
gpt_tokens = 0
gpt_costs = 0.0

logger = None

def enable_logging(model_logger):
    global logger
    logger = model_logger

def disable_logging():
    global logger
    logger = None

def get_gpt_tokens():
    return gpt_tokens

def get_embedding_tokens():
    return embedding_tokens

def get_gpt_costs():
    return gpt_costs

def get_embedding_costs():
    return embedding_costs

def get_total_cost():
    return embedding_costs + gpt_costs

def reset_total_cost():
    global embedding_costs, embedding_tokens, gpt_costs, gpt_tokens
    embedding_costs = 0.0
    embedding_tokens = 0
    gpt_costs = 0.0
    gpt_tokens = 0

def calculate_cost(model_name, usage):
    global embedding_costs, embedding_tokens, gpt_costs, gpt_tokens

    if model_name in MODELS:
        model = MODELS[model_name]
        token_count = usage['total_tokens']

        if model['model_type'] == MODEL_TYPE_LLM:
            gpt_tokens += token_count
            gpt_costs += (usage['prompt_tokens'] * model['input_price_per_token']) + (usage['completion_tokens'] * model['output_price_per_token'])
        else:
            embedding_tokens += token_count
            embedding_costs += token_count * model['price_per_token']
    else:
        print(f"Unknown model: {model_name}")


def get_max_tokens(model):
    return MODELS[model]['max_tokens']


def count_tokens(prompt, model):
    encoding = tiktoken.encoding_for_model(model)
    return len(encoding.encode(prompt))

def create_embedding(text, model="text-embedding-ada-002", connection_params={}):
    tries = 0
    while True:
        try:
            params = {
                'input': text if isinstance(text, list) else [text],
            }
            if 'deployment_id' not in connection_params:
                params['model'] = model

            response = openai.Embedding.create(
                **params,
                **connection_params
            )

            calculate_cost(model, response['usage'])

            embeddings = [d['embedding'] for d in response['data']]
            if len(embeddings) > 0:
                return embeddings
            else:
                return embeddings[0]

        except Exception as e:
            if tries >= 3:
                raise e
            
            print(e)
            time.sleep(5)
            tries += 1

def get_completion(
    prompt, 
    system_prompt='', 
    previous_messages=[], 
    functions=[], 
    model='gpt-4', 
    connection_params={}
):
    messages = __build_completion_messages(prompt, system_prompt=system_prompt, previous_messages=previous_messages)
    response = __get_completion(messages=messages, functions=functions, model=model, connection_params=connection_params)

    function = None
    message = response["choices"][0]["message"]
    content = message.get('content')
    if message.get("function_call"):
        function = {
            'name': message["function_call"]['name']
        }
        if 'arguments' in message["function_call"]:
            try:
                function['arguments'] = json.loads(message["function_call"]['arguments'])
            except json.decoder.JSONDecodeError:
                pass
    
    calculate_cost(model, response['usage'])
    __log_completion(
        prompt,
        system_prompt=system_prompt, 
        previous_messages=previous_messages, 
        functions=functions,
        response_content=content
    )

    return content, function


def get_streaming_completion(
    prompt, 
    system_prompt='', 
    previous_messages=[], 
    functions=[], 
    model='gpt-4', 
    connection_params={},
    content_stream_callback=None, 
    function_stream_callback=None
):
    connection_params['stream'] = True
    messages = __build_completion_messages(prompt, system_prompt=system_prompt, previous_messages=previous_messages)
    response = __get_completion(messages, functions=functions, model=model, connection_params=connection_params)

    content = ''
    function_name = ''
    function_arguments = ''

    for chunk in response:
        if len(chunk['choices']) == 0:
            continue
        
        choice = chunk['choices'][0]
        delta = choice.get('delta', {})
        
        if delta.get('content'):
            content += delta['content']
            if content_stream_callback:
                content_stream_callback(delta['content'])
        
        function_call = delta.get('function_call')
        if function_call:
            if function_call.get('name'):
                function_name += function_call['name']
            if function_call.get('arguments'):
                function_arguments += function_call['arguments']
            if function_stream_callback:
                function_stream_callback(function_name, function_arguments)

    function = None
    if function_name:
        function = { 'name': function_name }
        if function_arguments:
            try:
                function['arguments'] = json.loads(function_arguments)
            except json.decoder.JSONDecodeError:
                pass

    if function:
        completion_tokens = content + json.dumps(function)
    else:
        completion_tokens = content
    
    completion_token_count = count_tokens(completion_tokens, model)
    prompt_tokens = json.dumps(messages)
    prompt_token_count = count_tokens(prompt_tokens, model)
    
    calculate_cost(model, {
        'total_tokens': prompt_token_count + completion_token_count,
        'prompt_tokens': prompt_token_count,
        'completion_tokens': completion_token_count
    })

    __log_completion(
        prompt,
        system_prompt=system_prompt, 
        previous_messages=previous_messages, 
        functions=functions,
        response_content=content
    )

    return content, function


def __build_completion_messages(prompt, system_prompt='', previous_messages=[]):
    messages = []

    if system_prompt:
        messages.append({
            "role": "system",
            "content": system_prompt
        })
    
    if len(previous_messages) > 0:
        for m in previous_messages:
            messages.append({
                "role": "user",
                "content": m['prompt']
            })
            messages.append({
                "role": "assistant",
                "content": m['response']
            })

    messages.append({
        "role": "user",
        "content": prompt
    })

    return messages

def __log_completion(prompt, system_prompt='', previous_messages=[], functions=[], response_content='', response_function=None):
    global logger
    if not logger:
        return
    prompt_log = ''

    if len(functions) > 0:
        fs_str = json.dumps(functions)
        prompt_log += f"TOOLS:\n{fs_str}\n\n"

    if system_prompt:
        prompt_log += f"SYSTEM: {system_prompt}\n\n"
    
    if len(previous_messages) > 0:
        prompt_log += "PREVIOUS MESSAGES:\n"
        for m in previous_messages:
            prompt_log += f"{m}\n"
        prompt_log += '\n'

    prompt_log += f"PROMPT: {prompt}\n\n"
    logger.log(prompt_log, response_content, function=response_function)


def __get_completion(
    messages,
    functions=None, 
    model='gpt-4', 
    connection_params={}
):
    attempts = 0
    max_retries = 3
    while True:
        try:
            attempts += 1
            params = { 
                'messages': messages,
                'temperature': TEMPERATURE
            }

            if len(functions) > 0:
                params['functions'] = functions
            if 'engine' not in connection_params:
                params['model'] = model

            return openai.ChatCompletion.create(
                **params,
                **connection_params
            )
        except Exception as e:
            time.sleep(15)
            if attempts >= max_retries:
                raise e



import os
import psycopg2
from pydantic import BaseModel
from bondai.tools import Tool
from bondai.models.openai import OpenAILLM, MODEL_GPT35_TURBO_16K

class Parameters(BaseModel):
    question: str
    thought: str

TOOL_NAME = 'database_query_tool'
QUERY_SYSTEM_PROMPT = "You are a helpful question and answer assistant designed to answer questions about a database. Use the provided information to answer the user's QUESTION at the very end."
TOOL_DESCRIPTION = (
    "This tool allows you to ask a question and retrieve data from the user's database."
    "DO NOT ask the user for any additional information about the database."
    "All necessary information to connect to the database has already been captured."
    "Just specify your question using the 'question' parameter."
    "Your question will automatically be turned into SQL and the response will contain the resulting data."
)

PG_URI = os.environ.get('PG_URI')
PG_HOST = os.environ.get('PG_HOST')
PG_PORT = int(os.environ.get('PG_HOST', '5432'))
PG_USERNAME = os.environ.get('PG_USERNAME')
PG_PASSWORD = os.environ.get('PG_PASSWORD')
PG_DBNAME = os.environ.get('PG_DBNAME')

# PG_URI = ''
# PG_HOST = 'conifer-chat-db-prod.cmgu1yo5bmnj.us-west-2.rds.amazonaws.com'
# PG_PORT = int(os.environ.get('PG_HOST', '5432'))
# PG_USERNAME = 'conifer_chat_api_user'
# PG_PASSWORD = '7totx?97c$PJsLX9'
# PG_DBNAME = 'conifer_chat'

MAX_QUERY_RETRIES = 3

QUERY_PROMPT_TEMPLATE = """
Using the database schema below respond with a SQL query to answer the user's QUESTION.
VERY IMPORTANT: You must respond ONLY with a SQL query. Do not respond with any other text.

# DATABASE SCHEMA #

###DBSCHEMA###


# QUESTION #

###QUESTION###

VERY IMPORTANT: You must respond ONLY with a SQL query. Do not respond with any other text.
"""

RESPONSE_PROMPT_TEMPLATE = """
You were asked the following question:
###QUESTION###

The following information was returned from the database:
###QUERY_RESULTS###

Please respond with a friendly text response to the user's question.
"""

class DatabaseQueryTool(Tool):
    def __init__(self, 
                 pg_uri=PG_URI, 
                 pg_host=PG_HOST, 
                 pg_port=PG_PORT, 
                 pg_username=PG_USERNAME, 
                 pg_password=PG_PASSWORD, 
                 pg_dbname=PG_DBNAME,
                 llm=OpenAILLM(MODEL_GPT35_TURBO_16K)
            ):
        super(DatabaseQueryTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.pg_uri = pg_uri
        self.pg_host = pg_host
        self.pg_port = pg_port
        self.pg_username = pg_username
        self.pg_password = pg_password
        self.pg_dbname = pg_dbname
        self.llm = llm
    
    def run(self, arguments):
        question = arguments['question']

        if question is None:
            raise Exception('question is required')

        schema_str = self.__get_database_schema()
        query_prompt = QUERY_PROMPT_TEMPLATE.replace("###DBSCHEMA###", schema_str)
        query_prompt = query_prompt.replace("###QUESTION###", question)

        attempts = 0
        while True:
            try:
                query_response, _ = self.llm.get_completion(query_prompt)
                rows, colnames = self.__query_database(query_response)
                return self.__format_response(rows, colnames)
            except Exception as e:
                attempts += 1
                if attempts > MAX_QUERY_RETRIES:
                    raise e
    
    def __format_response(self, rows, colnames):
        markdown_output = "| " + " | ".join(colnames) + " |\n"
        markdown_output += "| " + " | ".join(["---"] * len(colnames)) + " |\n"

        for row in rows:
            markdown_output += "| " + " | ".join(map(str, row)) + " |\n"
        
        return markdown_output

    def __get_database_connection(self):
        if self.pg_uri:
            # Establish the connection
            return psycopg2.connect(self.pg_uri, sslmode='require')
        else:
            # Establish the connection
            return psycopg2.connect(
                host=self.pg_host,
                port=self.pg_port,
                user=self.pg_username,
                password=self.pg_password,
                dbname=self.pg_dbname
            )
    
    def __query_database(self, query):
        connection = None
        cursor = None
        try:
            connection = self.__get_database_connection()

            # Create a new cursor
            cursor = connection.cursor()

            # Execute the query
            cursor.execute(query)

            colnames = [desc[0] for desc in cursor.description]

            # Fetch all rows from the query result
            return cursor.fetchall(), colnames
        finally:
            # Close the cursor and the connection
            if cursor:
                cursor.close()
            if connection:
                connection.close()
    
    def __get_database_schema(self):
        # Query the schema from information_schema
        rows, _ = self.__query_database((
            "SELECT table_name, column_name, data_type, is_nullable, column_default\n"
            "FROM information_schema.columns\n"
            "WHERE table_schema = 'public'\n"
            "ORDER BY table_name, ordinal_position;"
        ))

        # Process the rows to create a formatted string
        schema_str = ""
        current_table = None
        for row in rows:
            table, column, data_type, is_nullable, default = row
            if table != current_table:
                if current_table:
                    schema_str += "\n"
                schema_str += f"Table: {table}\n"
                schema_str += "-" * (len(table) + 8) + "\n"
                current_table = table
            schema_str += f"{column}: {data_type}"
            if is_nullable == "NO":
                schema_str += " NOT NULL"
            if default:
                schema_str += f" DEFAULT {default}"
            schema_str += "\n"

        return schema_str


        






import os
from bondai.models import LLM
from .openai_wrapper import get_streaming_completion, get_completion, count_tokens, get_max_tokens
from .openai_connection_params import (
    GPT_35_CONNECTION_PARAMS,
    GPT_4_CONNECTION_PARAMS,
)
from .openai_models import (
    MODEL_NAMES, 
    MODELS, 
    MODEL_TYPE_LLM, 
    MODEL_FAMILY_GPT4
)

class OpenAILLM(LLM):
    def __init__(self, model, connection_details={}):
        if model not in MODEL_NAMES:
            raise Exception(f"Model {model} is not supported.")
        if MODELS[model]['model_type'] != MODEL_TYPE_LLM:
            raise Exception(f"Model {model} is not an LLM model.")
        self.model = model

    def supports_streaming(self):
        return True

    def get_completion(self, prompt, system_prompt='', previous_messages=[], functions=[]):
        if MODELS[self.model]['family'] == MODEL_FAMILY_GPT4:
            connection_params = GPT_4_CONNECTION_PARAMS
        else:
            connection_params = GPT_35_CONNECTION_PARAMS
        
        return get_streaming_completion(
            prompt, 
            system_prompt, 
            previous_messages, 
            functions, 
            self.model, 
            connection_params=connection_params)

    def get_streaming_completion(self, prompt, system_prompt='', previous_messages=[], functions=[], content_stream_callback=None, function_stream_callback=None):
        if MODELS[self.model]['family'] == MODEL_FAMILY_GPT4:
            connection_params = GPT_4_CONNECTION_PARAMS
        else:
            connection_params = GPT_35_CONNECTION_PARAMS
        
        return get_streaming_completion(
            prompt, 
            system_prompt, 
            previous_messages, 
            functions, 
            self.model, 
            connection_params=connection_params, 
            content_stream_callback=content_stream_callback,
            function_stream_callback=function_stream_callback)

    def count_tokens(self, prompt):
        return count_tokens(prompt, self.model)

    def get_max_tokens(self):
        return get_max_tokens(self.model)


#!/usr/bin/env python3

import os
import sys
import json
from termcolor import cprint, colored
from bondai import Agent, BudgetExceededException
from bondai.models.openai import enable_logging
from bondai.tools import HumanTool
from bondai.cli import get_tools, OnboardingTool

if not os.environ.get('OPENAI_API_KEY'):
    cprint(f"Please set the OPENAI_API_KEY environment variable.", 'red')
    exit(1)

cprint(f"Loading BondAI...", 'white')

enable_dangerous = "--enable_dangerous" in sys.argv
tool_options = get_tools()

if not enable_dangerous:
    tool_options = [tool for tool in tool_options if not tool.dangerous]

tool_descriptions = ''.join([f"{tool.name}: {tool.description}\n" for tool in tool_options])

WELCOME_TEXT = """
Welcome to BondAI!
I have been trained to help you with a variety of tasks.
To get started, just tell me what task you would like me to help you with.
If you would like to exit, just type 'exit'.

What can I help you with today?
"""

ONBOARDING_TASK = f"""
You are BondAI's helpful and friendly Onboarding AI assistant. You should communicate in a friendly and helpful manner.
Your job is to help understand what task the user wants to complete and gather all the information needed to start the task.
This information will then be given to another AI assistant that will complete the task.
Once you have gathered ALL of the required information from the user you will call the 'final_answer' tool to send the information to the next AI assistant.
ALWAYS greet the user with a friendly message.


You MUST gather the following REQUIRED information from the user:
- The Task Description (task_description): This must be detailed enough for the next AI assistant to understand what the user wants to do. Ask the user any necessary follow up questions. Confirm that the description you have with the user before calling the 'final_answer' tool.
- The Task Budget (task_budget): This is the maximum amount of money the user is willing to spend on this task. Ask the user any necessary follow up questions. Confirm that the budget you have with the user before calling the 'final_answer' tool.


IMPORTANT information:
- ALWAYS greet the user with a friendly message.
- If the user asks to exit you should call the 'final_answer' tool with the 'user_exit' parameter set to True. DO NOT ask for any other information from the user.
- ALWAYS confirm the Task Description and Task Budget with the user before calling the 'final_answer' tool.


REMEMBER: ALWAYS greet the user with a friendly message. Here is an example greeting you can use to get started:
```
{WELCOME_TEXT}
```

This is the list of all tools that you have access to:
{tool_descriptions}
"""

def get_task_definition():
    onboarding_tools = [
        HumanTool(),
        OnboardingTool()
    ]
    onboarding_agent = Agent(tools=onboarding_tools, quiet=True)
    onboarding_result = onboarding_agent.run(ONBOARDING_TASK)
    return json.loads(onboarding_result.output)

def run_task(task_config):
    task_description = task_config.get('task_description')
    task_budget = task_config.get('task_budget')

    print(colored("Starting Task:", 'white', attrs=["bold"]), colored(task_description, 'white'))
    print(colored("Budget:", 'white', attrs=["bold"]), colored(f"${task_budget}", 'white'))
    print(colored("Tools:", 'white', attrs=["bold"]), colored("\n".join(tool_ids), 'white'))
    
    agent = Agent(tools=tool_options, budget=task_budget, quiet=False)
    try:
        result = agent.run(task_description)
        cprint(f"\n\n{result.output}\n", 'white')
    except BudgetExceededException as e:
        cprint(f"\n\nThe budget for this task has been exceeded and will stop.\n", 'red')


def main():
    while True:
        task_config = get_task_definition()
        if task_config.get('user_exit'):
            break
        run_task(task_config)

if __name__ == '__main__':
    main()



import re
import os
import requests
import time
from pydantic import BaseModel
from bondai.tools import Tool

TOOL_NAME = 'phone_call_tool'
TOOL_DESCRIPTION = (
    "This tool interfaces allows you to make phone calls. The response you get back from this tool will be the call transcript.\n"
    "\nParameters:\n"
    "- phone_number (Required): The phone number of the person or company to call. For international numbers, "
    "include the country code without additional formatting (e.g., '+447700900077'). U.S. numbers may include "
    "formatting, but removing extra characters is recommended.\n"
    "- task (Required): This should be a highly detailed description of the task that should be performed on this call. Also, provide context about the interaction and give detailed instructions.\n"
    "- request_data (Optional, default={}): A dictionary that provides information for the AI during the call. Useful for giving the AI specific facts like the caller’s name, etc.\n"
)


BLAND_AI_API_KEY = os.environ.get('BLAND_AI_API_KEY')
BLAND_AI_VOICE_ID = int(os.environ.get('BLAND_AI_VOICE_ID', '0'))
BLAND_AI_CALL_TIMEOUT = int(os.environ.get('BLAND_AI_VOICE_ID', '300'))
API_ENDPOINT = 'https://api.bland.ai/'

# Interval for checking call status (in seconds)
CHECK_INTERVAL = 2


class CallParameters(BaseModel):
    phone_number: str
    task: str
    request_data: dict = {}
    thought: str




def validate_phone_number(phone):
    # International numbers (starting with '+')
    international_pattern = r'^\+\d{1,15}$'  # Starts with +, followed by 1 to 15 digits.

    # U.S. numbers (may include formatting characters)
    us_pattern = r'^(?:\+1)?[ -]?(\d{3})[ -]?(\d{3})[ -]?(\d{4})$'  # This considers formats like +1 123-456-7890, 123 456 7890, 123-456-7890.

    if re.match(international_pattern, phone):
        return True
    elif re.match(us_pattern, phone):
        return True
    else:
        error_msg = ("Invalid phone number format. "
                     "For international numbers, include the country code and exclude additional formatting. E.g. '+447700900077'. "
                     "For U.S. numbers, you may include formatting, but it's recommended to strip all additional characters.")
        raise ValueError(error_msg)


class BlandAITool(Tool):
    def __init__(self, bland_ai_api_key=BLAND_AI_API_KEY):
        super(BlandAITool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, CallParameters)
        self.bland_ai_api_key=bland_ai_api_key

    def run(self, arguments):
        if arguments.get('phone_number') is None:
            raise Exception("phone_number is required.")
        if arguments.get('task') is None:
            raise Exception("task is required.")

        validate_phone_number(arguments['phone_number'])

        # Start the call
        arguments['reduce_latency'] = False
        call_id = self.start_call(arguments)
        if not call_id:
            raise Exception("Failed to start the call.")
        
        # Monitor the call until completion
        start_time = time.time()
        while True:
            # Check for timeout
            elapsed_time = time.time() - start_time
            if elapsed_time > BLAND_AI_CALL_TIMEOUT:
                self.end_call(call_id)
                raise TimeoutError(f"Call exceeded the maximum allowed time of {BLAND_AI_CALL_TIMEOUT} seconds.")
            

            completed, transcripts = self.check_call_status(call_id)
            if completed:
                return f"Call to {arguments['phone_number']} has completed.\n\nTranscripts:\n{transcripts}"
            time.sleep(CHECK_INTERVAL)

    def start_call(self, arguments):
        headers = {'authorization': self.bland_ai_api_key}
        response = requests.post(API_ENDPOINT + 'call', json=arguments, headers=headers)
        if response.status_code == 200:
            resp_data = response.json()
            if resp_data["status"] == "success":
                return resp_data['call_id']
        return None

    def check_call_status(self, call_id):
        headers = {'authorization': self.bland_ai_api_key}
        data = {'call_id': call_id}
        response = requests.post(API_ENDPOINT + 'logs', json=data, headers=headers)
        
        if response.status_code == 200:
            resp_data = response.json()
            if resp_data['completed']:
                transcripts = "\n".join([f"{item['user']}: {item['text']}" for item in resp_data['transcripts']])
                return True, transcripts
        
        return False, None
    
    def end_call(self, call_id):
        headers = {'authorization': self.bland_ai_api_key}
        data = {'call_id': call_id}
        requests.post(API_ENDPOINT + 'end', json=data, headers=headers)




from .tool import Tool, EmptyParameters, InputParameters
from .human_tool import HumanTool
from .agent_tool import AgentTool
from .dalle_tool import DalleTool
from .python_repl_tool import PythonREPLTool
from .shell_tool import ShellTool
from .langchain_tool import LangChainTool
from .response_query import ResponseQueryTool

__all__ = [
    "Tool",
    "AgentTool",
    "DalleTool",
    "PythonREPLTool",
    "ShellTool",
    "EmptyParameters",
    "InputParameters",
    "HumanTool",
    "LangChainTool",
    "ResponseQueryTool"
]


import json
import requests
from socketio import Client

class BondAIAPIClient:
    def __init__(self, base_url='http://127.0.0.1:5000'):
        self.base_url = base_url
        self.ws_client = None
        self._events = {
            'agent_message': [],
            'agent_started': [],
            'agent_step_completed': [],
            'agent_completed': [],
        }
    
    def on(self, event_name):
        """Register a callback to an event."""
        if event_name not in self._events:
            raise ValueError(f"Unsupported event '{event_name}'")

        def decorator(callback):
            self._events[event_name].append(callback)
            return callback
        
        return decorator

    def _trigger_event(self, event_name, *args, **kwargs):
        """Trigger the specified event."""
        for callback in self._events.get(event_name, []):
            callback(*args, **kwargs)
    
    def connect_ws(self):
        if self.ws_client:
            self.disconnect_ws()
        self.ws_client = Client()
        self.ws_client.connect(self.base_url)

        @self.ws_client.on('message')
        def on_message(message):
            message = json.loads(message)
            if message.get('event') == 'agent_message':
                self._trigger_event('agent_message', message['data']['message'])
            elif message.get('event') == 'agent_step_completed':
                self._trigger_event('agent_step_completed', message['data']['step'])
            elif message.get('event') == 'agent_started':
                self._trigger_event('agent_started')
            elif message.get('event') == 'agent_completed':
                self._trigger_event('agent_completed')

    def disconnect_ws(self):
        if self.ws_client:
            self.ws_client.disconnect()
            self.ws_client = None
    
    def is_ws_connected(self):
        return self.ws_client and self.ws_client.connected

    def send_ws_message(self, event, data):
        if not self.is_ws_connected():
            self.connect_ws()
        message = {"event": event, "data": data}
        message_bytes = json.dumps(message).encode('utf-8')
        self.ws_client.send(message_bytes)
    
    def send_user_message(self, message):
        self.send_ws_message("user_message", {"message": message})

    def _request(self, method, endpoint, data=None):
        url = f"{self.base_url}{endpoint}"
        try:
            if method == 'GET':
                response = requests.get(url)
            elif method == 'POST':
                response = requests.post(url, json=data)
            elif method == 'DELETE':
                response = requests.delete(url, json=data)
            else:
                raise ValueError(f"Unsupported method: {method}")

            response.raise_for_status()
            return response.json()
        except requests.RequestException as e:
            raise Exception(f"HTTP Request Error: {e}")

    def get_agent(self):
        return self._request('GET', '/agent')

    def start_agent(self, task=None, task_budget=None, max_steps=None):
        data = {}
        if task:
            data['task'] = task
        if task_budget:
            data['task_budget'] = task_budget
        if max_steps:
            data['max_steps'] = max_steps
        return self._request('POST', '/agent/start', data)

    def add_agent_tool(self, tool_name):
        data = {'tool_name': tool_name}
        return self._request('POST', '/agent/tools', data)

    def remove_agent_tool(self, tool_name):
        return self._request('DELETE', f'/agent/tools/{tool_name}')

    def get_tools(self):
        return self._request('GET', '/tools')